//
// Copyright (C) 2008 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#ifndef __IP_PACKET_H
#define __IP_PACKET_H

#include "Packet.h"

//
// constants for the encapsulated transport protocols
//

#define IPTYPE_IP	0x04
#define IPTYPE_TCP	0x06
#define IPTYPE_UDP	0x11
#define IPTYPE_ICMP	0x01
#define IPTYPE_IPV6	0x29

//
// some ip constants
//

#define	IP_ADDR_LEN			4
#define IP_HEADER_NO_OPTIONS_LEN	20

//
// masks to fetch combined values
// 

#define IP_HEADER_VERSION(x)		((x &	0xF0	) >>  4	)
#define IP_HEADER_LEN(x)		((x &	0x0F	) *   4	)
#define IP_HEADER_FLAGS(x)		((x &	0xE000	) >> 13	)
#define IP_HEADER_FRAGOFFSET(x)		((x &	0x1FFF	)	)

//
// the IP header
// 

#pragma pack (1)
typedef struct _IP_HEADER {
	unsigned char		version_len;		// 4 bit version and 4 bit length 
	unsigned char		typeofservice;		// type of service . configuration item IpTos
	unsigned short		totallen;		// total length
	unsigned short		identification;		// identification. configuration item IpIdent
	unsigned short		flags_fragoffset;	// 3  bit flags and 13 bit fragment offset field. configuration item IpFragoffset and IpFlags
	unsigned char		ttl;			// time to live. configuration item IpTtl				
	unsigned char		protocol;		// protocol. configuration item IpSourceip
	unsigned short		checksum;		// checksum
	unsigned long		sourceip;		// source ip address. configuration item IpSource
	unsigned long		destip;			// dest ip address. configuration item IpDestip
} IP_HEADER, *PIP_HEADER;
#pragma pack ()

//
// a IP address
//

typedef struct _IP_ADDR {
	unsigned char x [IP_ADDR_LEN];

	string toString () {
		ostringstream out;

		for (int i=0; i<IP_ADDR_LEN; i++) {
			out << (int)x[i];
			if (i < IP_ADDR_LEN-1) out << ".";
		}

		return out.str ();
	}
} IP_ADDR, *PIP_ADDR;

//
// IP packet class
//

class IpPacket : public Packet
{
public:
	IpPacket(void);
	~IpPacket(void);

	bool			parsePacket		();
	void			assemblePacket		();
	string			toString		();
	unsigned int		getMinProtocolSize	();

	static Packet::PROTOCOL	getTransportProtocol	(unsigned char proto);

	unsigned char		getVersion		();
	unsigned char		getHeaderlength		();
	unsigned char		getTos			();
	unsigned short		getTotallen		();
	unsigned short		getIdentification	();
	unsigned short		getFlags		();
	unsigned short		getFragoffset		();
	unsigned char		getTtl			();
	unsigned char		getProtocol		();
	unsigned short		getChecksum		();
	IP_ADDR			getSourceip		();
	IP_ADDR			getDestip		();
	Packet::PAYLOAD_BUFFER	getOptions		();

	void			setVersion		(unsigned char	ver);
	void			setHeaderlength		(unsigned char	len);
	void			setTos			(unsigned char	tos);
	void			setTotallen		(unsigned short	totlen);
	void			setIdentification	(unsigned short	ident);
	void			setFlags		(unsigned short	flgs);
	void			setFragoffset		(unsigned short	fragoff);
	void			setTtl			(unsigned char	tl);
	void			setProtocol		(unsigned char	proto);
	void			setChecksum		(unsigned short checks);
	void			setSourceip		(IP_ADDR sip);
	void			setDestip		(IP_ADDR dip);
	void			setOptions		(unsigned char* buf, int len);

	static AnonPrimitive*	anonTos;
	static AnonPrimitive*	anonIdentifier;
	static AnonPrimitive*	anonFlags;
	static AnonPrimitive*	anonFragoffset;
	static AnonPrimitive*	anonTtl;
	static AnonPrimitive*	anonSourceip;
	static AnonPrimitive*	anonDestip;
	static AnonPrimitive*	anonOptions;

private:

	IP_HEADER		header;
	Packet::PAYLOAD_BUFFER	options; // configuration item IpOptions
	
protected:

	static unsigned short	checksum		(PIP_HEADER ipheader);

};

#endif // __IP_PACKET_H
